/*
 *  ircd-hybrid: an advanced Internet Relay Chat Daemon(ircd).
 *  m_mode.c: Sets a user or channel mode.
 *
 *  Copyright (C) 2002 by the past and present ircd coders, and others.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 *  USA
 *
 *  $Id: m_mode.c,v 1.4 2002/08/29 23:43:08 strtok Exp $
 */

#include "stdinc.h"
#include "tools.h"
#include "handlers.h"
#include "channel.h"
#include "channel_mode.h"
#include "vchannel.h"
#include "client.h"
#include "hash.h"
#include "irc_string.h"
#include "ircd.h"
#include "numeric.h"
#include "s_user.h"
#include "s_conf.h"
#include "s_serv.h"
#include "send.h"
#include "msg.h"
#include "parse.h"
#include "modules.h"
#include "packet.h"

static void m_mode(struct Client*, struct Client*, int, char**);

struct Message mode_msgtab = {
  "MODE", 0, 0, 0, 0, MFLG_SLOW, 0,
  {m_unregistered, m_mode, m_mode, m_mode}
};
#ifndef STATIC_MODULES

void
_modinit(void)
{
  mod_add_cmd(&mode_msgtab);
}

void
_moddeinit(void)
{
  mod_del_cmd(&mode_msgtab);
}


const char *_version = "$Revision: 1.4 $";
#endif

static char *mode_help[] = {
	    "\2MODE \37channel\37 \37[+-bcehiIklmMnopsRtv]\37 \37[args]\37\2",
	    " ",
	    "Sets specified modes on \37channel\37.  Modes \37ovbkleIh\37 all",
	    "require arguments.",
	    "Channel modes have the following meanings:",
	    " ",
            " a        Anonymous ops, channel ops are hidden.",
	    " b \37mask\37   Ban clients matching \37mask\37 from the channel",
	    " c        No colored messages to the channel",
            " e \37mask\37   Exempt clients matching \37mask\37 from any bans",
            " h \37nick\37   Give channel halfop privleges to \37nick\37",
	    " i        Channel is set invite-only",
            " I \37mask\37   Exempt clients matching \37mask\37 from +i",
	    " k \37key\37    Channel key becomes \37key\37; must be specified",
	    "          to enter the channel",
	    " l \37limit\37  There may be no more than \37limit\37 users in",
	    "          the channel at once",
	    " m        Channel is moderated; only operators and",
	    "          voices may speak",
	    " M        Only those identified to their registered nick may",
	    "          speak.",
	    " n        No messages from users outside the channel",
	    " o \37nick\37   Give channel operator privileges to \37nick\37",
	    " p        Channel is private; knocks will not be allowed and",
            "          halfops cannot set +/-h.",
	    " s        Channel is secret; will not appear on /list or",
	    "          /whois",
	    " R        Only users with registered nicknames may join",
	    " t        Only operators may change the channel's topic",
	    " v \37nick\37   Give \37nick\37 a voice in the channel",
	    " ",
	    "\2MODE \37nick\37 \37[+-giprRsw]\37\2",
	    " ",
	    "Sets specified modes on \37user\37.  \37user\37 is always",
	    "yourself.",
	    "User modes have the following meanings:",
	    " ",
            " g        Caller ID. A server side ignore feature. Only users in",
            "          /accept list can message you.",
	    " i        Invisible; other users cannot match you with /who unless",
	    "          they share a channel with you",
	    " p        Only show channels in your /whois to those who share the",
	    "          channels with you",
	    " r        Nick is registered and identified to services.  Only",
	    "          services may set or unset this mode",
	    " R        Only accept private messages from users who are",
            "          identified to their registered nick",
	    " s        See server notices",
	    " w        See wallops",
	    " ",
	    NULL
    };


static char *mode_oper_help[] = {
	    "Operator only user modes:",
            " ",
            " a        Administrator status",
            " b        See bot and drone flooding notices",
            " c        See client connection/quit notices",
            " d        See debug notices",
            " f        See auth {} full notices",
            " j        See rejected client notices",
            " k        See server kill notices",
            " l        See locops notices",
            " n        See client nick changes",
            " o        Operator status",
            " u        See unauthorized client messages",
            " x        See remote server connection and split notices",
            " y        See LINKS, STATS and TRACE notices",
            " z        See operator generated wallops",
	    " ",
	    NULL
    };



/*
 * m_mode - MODE command handler
 * parv[0] - sender
 * parv[1] - channel
 */
static void m_mode(struct Client *client_p, struct Client *source_p,
              int parc, char *parv[])
{
  struct Channel* chptr=NULL;
  struct Channel* root;
  static char     modebuf[MODEBUFLEN];
  static char     parabuf[MODEBUFLEN];
  dlink_node	*ptr;
  int n = 2;
  char **help;
#ifdef VCHANS
  struct Channel* vchan;
#endif

  /* No parameters given? Give client a list of 
     possible parameters (+blitzed) */
  if(parc == 0)
    {
       for(help = mode_help; *help; *help++)
          sendto_one(source_p, form_str(RPL_COMMANDSYNTAX), me.name,
                     source_p->name, *help);
       
       if(IsOper(source_p))
          for(help = mode_oper_help; *help; *help++)
             sendto_one(source_p, form_str(RPL_COMMANDSYNTAX), me.name,
                        source_p->name, *help);
       return;
    }

  /* Now, try to find the channel in question */
  if (!IsChanPrefix(parv[1][0]))
    {
      /* if here, it has to be a non-channel name */
      user_mode(client_p, source_p, parc, parv);
      return;
    }

  if (!check_channel_name(parv[1]))
    { 
      sendto_one(source_p, form_str(ERR_BADCHANNAME),
		 me.name, parv[0], (unsigned char *)parv[1]);
      return;
    }
	  
  chptr = hash_find_channel(parv[1]);

  if (chptr == NULL)
    {
      /* if chptr isn't found locally, it =could= exist
       * on the uplink. So ask.
       */
      
      /* LazyLinks */
      /* this was segfaulting if we had no servers linked.
       *  -pro
       */
      /* only send a mode upstream if a local client sent this request
       * -davidt
       */
      if (MyClient(source_p) && !ServerInfo.hub && uplink &&
	   IsCapable(uplink, CAP_LL))
	{
#if 0
	  /* cache the channel if it exists on uplink
	   * If the channel as seen by the uplink, has vchans,
	   * the uplink will have to SJOIN all of those.
	   */
	  /* Lets not for now -db */

	  sendto_one(uplink, ":%s CBURST %s",
                     me.name, parv[1]);
	  
#endif
	  sendto_one(uplink, ":%s MODE %s %s",
		     source_p->name, parv[1], (parv[2] ? parv[2] : ""));
	  return;
	}
      else
	{
	  sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL),
		     me.name, parv[0], parv[1]);
	  return;
	}
    }
  
  /* Now known the channel exists */

  root = chptr;

#ifdef VCHANS
  if ((parc > 2) && parv[2][0] == '!')
    {
     struct Client *target_p;
     if (!(target_p = find_client(++parv[2])))
       {
        sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name,
                   parv[0], root->chname);
        return;
       }
     if ((chptr = map_vchan(root, target_p)) == NULL)
       {
        sendto_one(source_p, form_str(ERR_NOSUCHCHANNEL), me.name,
                   parv[0], root->chname);
        return;
       }
     n++;
    }

  else
    {
      if (HasVchans(chptr))
        {
          if ((vchan = map_vchan(chptr,source_p)) != NULL)
            chptr = vchan; /* root = chptr, chptr = vchan */

          /* XXX - else? the user isn't on any vchan, so we
           *       end up giving them the mode of the root
           *       channel.  MODE #vchan !nick ? (ugh)
           */
        }
      else if (IsVchan(chptr))
        {
          vchan = find_bchan(chptr);
          root = vchan;  /* root = vchan, chptr = chptr */

          /* XXX - else? the user isn't on any vchan,
           *       but they asked for MODE ##vchan_12345
           *       we send MODE #vchan
           */
        }
    }
#endif

  if (parc < n+1)
    {
      channel_modes(chptr, source_p, modebuf, parabuf);
      sendto_one(source_p, form_str(RPL_CHANNELMODEIS),
		 me.name, parv[0], parv[1],
		 modebuf, parabuf);
      
      /* Let opers see the "true" TS everyone else see's
       * the top root chan TS
       */
      if (!IsOper(source_p))
	sendto_one(source_p, form_str(RPL_CREATIONTIME),
		   me.name, parv[0],
		   parv[1], root->channelts);
      else
	sendto_one(source_p, form_str(RPL_CREATIONTIME),
		   me.name, parv[0],
		   parv[1], chptr->channelts);
    }
  /* bounce all modes from people we deop on sjoin */
  else if((ptr = find_user_link(&chptr->deopped, source_p)) == NULL)
  {
    /* Finish the flood grace period... */
    if(MyClient(source_p) && !IsFloodDone(source_p))
    {
      if((parc == n) && (parv[n-1][0] == 'b') && (parv[n-1][1] == '\0'))
        ;
      else
        flood_endgrace(source_p);
    }

    set_channel_mode(client_p, source_p, chptr, parc - n, parv + n, 
                     root->chname);
  }
}




